package fucoin.configurations;

import akka.actor.ActorRef;
import akka.actor.Props;
import akka.pattern.Patterns;
import akka.util.Timeout;
import fucoin.AbstractNode;
import fucoin.actions.control.ActionAnnounceWalletCreation;
import fucoin.actions.join.ActionTellSupervisor;
import fucoin.configurations.internal.ConfigurationCreator;
import fucoin.supervisor.SuperVisorImpl;
import fucoin.wallet.WalletImpl;
import scala.concurrent.Await;
import scala.concurrent.Future;
import scala.concurrent.duration.Duration;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;

/**
 *
 */
public abstract class AbstractConfiguration extends AbstractNode {

    protected ActorRef superVisor;

    protected final List<ActorRef> activeActors = new ArrayList<>();

    protected Timeout timeout = new Timeout(Duration.create(10, "seconds"));

    public static Props props(Class configurationClass) {

        return Props.create(new ConfigurationCreator(configurationClass));
    }

    /**
     * Spawns a new wallet and blocks until it has received its initial money
     * @throws Exception on timeout
     */
    public ActorRef spawnWallet(String name, boolean createGUI) throws Exception {
        Future<Object> future = Patterns.ask(superVisor, new ActionAnnounceWalletCreation(1, self()), timeout);
        ActorRef wallet = createWallet(name, createGUI);
        Await.result(future, timeout.duration());
        return wallet;
    }

    /**
     * Creates a wallet without blocking until the wallet was created
     */
    public ActorRef createWallet(String name, boolean createGUI) {
        Props props;
        int numOfWallets = activeActors.size();
        if (numOfWallets == 0) {
            props = WalletImpl.props(null, name, createGUI);
        } else {
            props = WalletImpl.props(activeActors.get(numOfWallets - 1), name, createGUI);
        }

        ActorRef actorRef = context().actorOf(props, name);

        activeActors.add(actorRef);

        if (numOfWallets == 0) {
            actorRef.tell(new ActionTellSupervisor(superVisor), superVisor);
        }

        return actorRef;
    }

    /**
     * Spawn multiple wallets and wait until they all have their initial FUC
     * @throws Exception on timeout
     */
    public void spawnWallets(int n, boolean createGUI) throws Exception {
        Future<Object> future = Patterns.ask(superVisor, new ActionAnnounceWalletCreation(n, self()), timeout);
        for (int i = 0; i < n; i++) {
            String nameOfTheWallet = "Wallet" + String.valueOf(activeActors.size());
            createWallet(nameOfTheWallet, createGUI);
        }
        Await.result(future, timeout.duration());
    }

    /**
     * Fetch a random wallet
     * @return
     */
    public ActorRef getRandomWallet() {
        return activeActors.get(ThreadLocalRandom.current().nextInt(activeActors.size()));
    }

    public List<ActorRef> wallets() {
        return this.activeActors;
    }

    /**
     * Create the supervisor node
     * @return
     */
    public ActorRef initSupervisor() {
        superVisor = context().actorOf(SuperVisorImpl.props(), "SuperVisorImpl");

        // Don't ask.
        try {
            Thread.sleep(200);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return superVisor;
    }

    @Override
    public void preStart() throws Exception {
        super.preStart();

        this.run();
    }

    public abstract void run();
}
